from dataclasses import dataclass, field
from typing import Dict, List

from netspresso.enums.metadata import TaskType
from netspresso.metadata.common import AvailableOption, BaseMetadata, InputShape


@dataclass
class ModelInfo:
    task: str = ""
    model: str = ""
    dataset: str = ""
    input_shapes: List[InputShape] = field(default_factory=lambda: [InputShape()])


@dataclass
class TrainingInfo:
    epochs: int = 0
    batch_size: int = 0
    learning_rate: float = 0.001
    optimizer: str = "Adam"
    scheduler: str = "CosineAnnealingWarmRestartsWithCustomWarmUp"


@dataclass
class TrainerMetadata(BaseMetadata):
    task_type: TaskType = TaskType.TRAIN
    output_dir: str = ""
    best_fx_model_path: str = ""
    best_onnx_model_path: str = ""
    hparams: str = ""
    runtime: str = ""
    model_info: ModelInfo = field(default_factory=ModelInfo)
    training_info: TrainingInfo = field(default_factory=TrainingInfo)
    training_result: Dict = field(default_factory=dict)
    available_options: List[AvailableOption] = field(default_factory=list)

    def update_model_info(self, task, model, dataset, input_shapes):
        self.model_info.task = task
        self.model_info.model = model
        self.model_info.dataset = dataset
        self.model_info.input_shapes = input_shapes

    def update_training_info(self, epochs, batch_size, learning_rate, optimizer, scheduler):
        self.training_info.epochs = epochs
        self.training_info.batch_size = batch_size
        self.training_info.learning_rate = learning_rate
        self.training_info.optimizer = optimizer
        self.training_info.scheduler = scheduler

    def update_training_result(self, training_summary):
        self.training_result = training_summary

    def update_output_dir(self, output_dir):
        self.output_dir = output_dir

    def update_best_fx_model_path(self, best_fx_model_path):
        self.best_fx_model_path = best_fx_model_path

    def update_best_onnx_model_path(self, best_onnx_model_path):
        self.best_onnx_model_path = best_onnx_model_path

    def update_hparams(self, hparams):
        self.hparams = hparams

    def update_available_options(self, available_options):
        self.available_options = [available_option.to() for available_option in available_options]



import dataclasses
from abc import ABC, abstractmethod

from netspresso.clients.launcher.v2.schemas import ResponseItem, ResponseItems


class TaskInterface(ABC):
    @abstractmethod
    def start(
        self, request_body: dataclasses, headers: dataclasses, file: dataclasses
    ) -> ResponseItem:
        """

        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def cancel(self, headers: dataclasses, task_id: str) -> ResponseItem:
        """

        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def read(self, headers: dataclasses, task_id: str) -> ResponseItem:
        """

        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def status(self, headers: dataclasses, task_id: str) -> ResponseItem:
        """

        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def delete(self, headers: dataclasses, task_id: str) -> ResponseItem:
        """

        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def options(self, headers: dataclasses) -> ResponseItems:
        """

        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def options_by_model_framework(self, headers: dataclasses, model_framework: str) -> ResponseItems:
        """

        :return:
        """
        raise NotImplementedError




from .request_body import RequestBenchmark
from .response_body import (
    ResponseBenchmarkFrameworkOptionItems,
    ResponseBenchmarkOptionItems,
    ResponseBenchmarkStatusItem,
    ResponseBenchmarkTaskItem,
)

__all__ = [
    RequestBenchmark,
    ResponseBenchmarkTaskItem,
    ResponseBenchmarkOptionItems,
    ResponseBenchmarkStatusItem,
    ResponseBenchmarkFrameworkOptionItems
]



from .common import (
    AuthorizationHeader,
    DeviceInfo,
    ModelOption,
    RequestPagination,
    ResponseItem,
    ResponseItems,
    ResponsePaginationItems,
    UploadFile,
)
from .compression import (
    Options,
    RecommendationOptions,
    RequestAutomaticCompressionParams,
    RequestAvailableLayers,
    RequestCreateCompression,
    RequestCreateRecommendation,
    RequestUpdateCompression,
    ResponseCompression,
    ResponseCompressionItem,
    ResponseCompressionItems,
    ResponseRecommendationItem,
    ResponseSelectMethod,
    ResponseSelectMethodItem,
)
from .model import (
    InputLayer,
    ModelBase,
    RequestCreateModel,
    RequestUploadModel,
    RequestValidateModel,
    ResponseModelItem,
    ResponseModelItems,
    ResponseModelOptions,
    ResponseModelStatus,
    ResponseModelUrl,
)



from netspresso.clients.utils.requester import Requester


class ExperimentAPI:
    def __init__(self, url: str):
        self.url = url

    def get_experiments(
        self,
        user_id,
        headers,
        skip=None,
        size=None,
        sort=None,
        name=None,
        type=None,
        network_arch=None,
        read_only=None,
        user_only=None,
    ):
        endpoint = f"{self.url}/users/{user_id}/experiments"
        params = {
            key: value
            for key, value in {
                "skip": skip,
                "size": size,
                "sort": sort,
                "name": name,
                "type": type,
                "network_arch": network_arch,
                "read_only": read_only,
                "user_only": user_only,
            }.items()
            if value is not None
        }

        response = Requester.get(url=endpoint, params=params, headers=headers)

        return response.json()

    def create_experiments(self, user_id, request_body, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments"

        response = Requester.post_as_json(url=endpoint, request_body=request_body, headers=headers)

        return response.json()

    def delete_experiment(self, user_id, experiment_id, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}"

        response = Requester.delete(url=endpoint, headers=headers)

        return response.json()

    def get_experiment(self, user_id, experiment_id, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}"

        response = Requester.get(url=endpoint, headers=headers)

        return response.json()

    def partial_update_experiment(self, user_id, experiment_id, request_body, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}"

        response = Requester.patch(url=endpoint, request_body=request_body, headers=headers)

        return response.json()

    def update_experiment(self, user_id, experiment_id, request_body, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}"

        response = Requester.put(url=endpoint, request_body=request_body, headers=headers)

        return response.json()

    def get_experiment_jobs(self, user_id, experiment_id, headers, skip=None, size=None, sort=None):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}/jobs"
        params = {
            key: value
            for key, value in {
                "skip": skip,
                "size": size,
                "sort": sort,
            }.items()
            if value is not None
        }

        response = Requester.get(url=endpoint, params=params, headers=headers)

        return response.json()

    def run_experiment_jobs(self, user_id, experiment_id, request_body, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}/jobs"

        response = Requester.post_as_json(url=endpoint, request_body=request_body, headers=headers)

        return response.json()

    def delete_experiment_job(self, user_id, experiment_id, job_id, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}/jobs/{job_id}"

        response = Requester.delete(url=endpoint, headers=headers)

        return response.json()

    def get_experiment_job(self, user_id, experiment_id, job_id, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}/jobs/{job_id}"

        response = Requester.get(url=endpoint, headers=headers)

        return response.json()

    def cancel_experiment_job(self, user_id, experiment_id, job_id, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}/jobs/{job_id}:cancel"

        response = Requester.post_as_json(url=endpoint, headers=headers)

        return response.json()

    def download_job_artifacts(self, user_id, experiment_id, job_id, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}/jobs/{job_id}:download"

        response = Requester.get(url=endpoint, headers=headers)

        return response.json()

    def resume_experiment_job(self, user_id, experiment_id, job_id, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}/jobs/{job_id}:resume"

        response = Requester.post_as_json(url=endpoint, headers=headers)

        return response.json()

    def get_specs_schema(self, user_id, experiment_id, action, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}/specs/{action}/schema"

        response = Requester.get(url=endpoint, headers=headers)

        return response.json()

    def get_list_files(self, user_id, experiment_id, job_id, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}/jobs/{job_id}:list_files"

        response = Requester.get(url=endpoint, headers=headers)

        return response.json()

    def download_selective_files(self, user_id, experiment_id, job_id, file_lists, best_model, latest_model, headers):
        endpoint = f"{self.url}/users/{user_id}/experiments/{experiment_id}/jobs/{job_id}:download_selective_files"
        params = {"file_lists": file_lists, "best_model": best_model, "latest_model": latest_model}

        response = Requester.get(url=endpoint, params=params, headers=headers, stream=True)

        return response



import time
from pathlib import Path
from typing import Optional, Union

from loguru import logger

from netspresso.base import NetsPressoBase
from netspresso.clients.auth import TokenHandler
from netspresso.clients.auth.response_body import UserResponse
from netspresso.clients.launcher import launcher_client_v2
from netspresso.clients.launcher.v2.schemas.task.benchmark.response_body import (
    BenchmarkTask,
)
from netspresso.enums import Status, TaskStatusForDisplay
from netspresso.enums.credit import ServiceTask
from netspresso.enums.device import (
    DeviceName,
    HardwareType,
    SoftwareVersion,
)
from netspresso.metadata.benchmarker import BenchmarkerMetadata
from netspresso.utils import FileHandler
from netspresso.utils.metadata import MetadataHandler


class BenchmarkerV2(NetsPressoBase):
    def __init__(self, token_handler: TokenHandler, user_info: UserResponse) -> None:
        """Initialize the Benchmarker."""

        super().__init__(token_handler)
        self.user_info = user_info

    def initialize_metadata(self, input_model_path: str):
        def create_metadata_with_status(status, error_message=None):
            metadata = BenchmarkerMetadata()
            metadata.status = status
            if error_message:
                logger.error(error_message)
            return metadata

        try:
            metadata = BenchmarkerMetadata()
        except Exception as e:
            error_message = f"An unexpected error occurred during metadata initialization: {e}"
            metadata = create_metadata_with_status(Status.ERROR, error_message)
        except KeyboardInterrupt:
            warning_message = "Benchmark task was interrupted by the user."
            metadata = create_metadata_with_status(Status.STOPPED, warning_message)
        finally:
            # Load existing metadata if available
            metadatas = []
            output_dir = Path(input_model_path).parent
            file_path = output_dir / "benchmark.json"
            if FileHandler.check_exists(file_path):
                metadatas = MetadataHandler.load_json(file_path)

            metadata.input_model_path = Path(input_model_path).resolve().as_posix()
            metadatas.append(metadata)
            MetadataHandler.save_benchmark_result(data=metadatas, folder_path=output_dir)

        return metadatas

    def benchmark_model(
        self,
        input_model_path: str,
        target_device_name: DeviceName,
        target_software_version: Optional[Union[str, SoftwareVersion]] = None,
        target_hardware_type: Optional[Union[str, HardwareType]] = None,
        wait_until_done: bool = True,
        sleep_interval: int = 30,
    ) -> BenchmarkerMetadata:
        """Benchmark the specified model on the specified device.

        Args:
            input_model_path (str): The file path where the model is located.
            target_device_name (DeviceName): Target device name.
            target_software_version (Union[str, SoftwareVersion], optional): Target software version. Required if target_device_name is one of the Jetson devices.
            target_hardware_type (Union[str, HardwareType], optional): Hardware type. Acceleration options for processing the model inference.
            wait_until_done (bool): If True, wait for the conversion result before returning the function.
                                If False, request the conversion and return the function immediately.

        Raises:
            e: If an error occurs during the benchmarking of the model.

        Returns:
            BenchmarkerMetadata: Benchmark metadata.
        """

        FileHandler.check_input_model_path(input_model_path)
        metadatas = self.initialize_metadata(input_model_path=input_model_path)

        try:
            metadata: BenchmarkerMetadata = metadatas[-1]
            output_dir = Path(input_model_path).parent
            if metadata.status in [Status.ERROR, Status.STOPPED]:
                return metadata

            self.validate_token_and_check_credit(service_task=ServiceTask.MODEL_BENCHMARK)

            # Get presigned_model_upload_url
            presigned_url_response = launcher_client_v2.benchmarker.presigned_model_upload_url(
                access_token=self.token_handler.tokens.access_token,
                input_model_path=input_model_path,
            )

            # Upload model_file
            launcher_client_v2.benchmarker.upload_model_file(
                access_token=self.token_handler.tokens.access_token,
                input_model_path=input_model_path,
                presigned_upload_url=presigned_url_response.data.presigned_upload_url,
            )

            # Validate model_file
            validate_model_response = launcher_client_v2.benchmarker.validate_model_file(
                access_token=self.token_handler.tokens.access_token,
                input_model_path=input_model_path,
                ai_model_id=presigned_url_response.data.ai_model_id,
            )

            # Start benchmark task
            benchmark_response = launcher_client_v2.benchmarker.start_task(
                access_token=self.token_handler.tokens.access_token,
                input_model_id=presigned_url_response.data.ai_model_id,
                data_type=validate_model_response.data.detail.data_type,
                target_device_name=target_device_name,
                hardware_type=target_hardware_type,
                input_layer=validate_model_response.data.detail.input_layers[0],
                software_version=target_software_version,
            )

            metadata.benchmark_task_info = benchmark_response.data.to()
            MetadataHandler.save_benchmark_result(data=metadatas, folder_path=output_dir)

            if wait_until_done:
                while True:
                    self.token_handler.validate_token()
                    benchmark_response = launcher_client_v2.benchmarker.read_task(
                        access_token=self.token_handler.tokens.access_token,
                        task_id=benchmark_response.data.benchmark_task_id,
                    )
                    if benchmark_response.data.status in [
                        TaskStatusForDisplay.FINISHED,
                        TaskStatusForDisplay.ERROR,
                        TaskStatusForDisplay.TIMEOUT,
                    ]:
                        break

                    time.sleep(sleep_interval)

            if benchmark_response.data.status == TaskStatusForDisplay.FINISHED:
                self.print_remaining_credit(service_task=ServiceTask.MODEL_BENCHMARK)
                metadata.status = Status.COMPLETED
                metadata.benchmark_result = benchmark_response.data.benchmark_result.to(
                    file_size=validate_model_response.data.file_size_in_mb
                )
                logger.info("Benchmark task was completed successfully.")
            else:
                metadata = self.handle_error(metadata, ServiceTask.MODEL_BENCHMARK, benchmark_response.data.error_log)

        except Exception as e:
            metadata = self.handle_error(metadata, ServiceTask.MODEL_BENCHMARK, e.args[0])
        except KeyboardInterrupt:
            metadata = self.handle_stop(metadata, ServiceTask.MODEL_BENCHMARK)
        finally:
            metadatas[-1] = metadata
            MetadataHandler.save_benchmark_result(data=metadatas, folder_path=output_dir)

        return metadata

    def get_benchmark_task(self, benchmark_task_id: str) -> BenchmarkTask:
        """Get information about the specified benchmark task using the benchmark task UUID.

        Args:
            benchmark_task_id (str): Benchmark task UUID of the benchmark task.

        Raises:
            e: If an error occurs while retrieving information about the benchmark task.

        Returns:
            BenchmarkTask: Model benchmark task object.
        """

        self.token_handler.validate_token()

        response = launcher_client_v2.benchmarker.read_task(
            access_token=self.token_handler.tokens.access_token,
            task_id=benchmark_task_id,
        )
        return response.data



import cv2
import numpy as np
from matplotlib import pyplot as plt

from netspresso.inferencer.visualizers.utils import voc_color_map


class DetectionVisualizer:
    def __init__(self, class_map=None, normalized=False, brightness_factor=1.5):
        self.cmap = voc_color_map(N=256, normalized=normalized, brightness_factor=brightness_factor)
        self.class_map = class_map

    def draw(self, image, pred, model_input_shape=None, text_scale=0.7, text_thickness=2, bbox_thickness=2):
        resize_factor = max((image.shape[0] / model_input_shape[0]), (image.shape[1] / model_input_shape[1]))
        pred[0][:, 1::2] *= resize_factor
        pred[0][:, :4:2] *= resize_factor

        visualize_image = image.copy()
        for bbox_label, class_label in zip(pred[0], pred[1]):
            # Get bbox coordinates
            x1 = int(bbox_label[0])
            y1 = int(bbox_label[1])
            x2 = int(bbox_label[2])
            y2 = int(bbox_label[3])

            # Get bbox color
            color = self.cmap[class_label].tolist()

            # Draw bbox
            visualize_image = cv2.rectangle(visualize_image, (x1, y1), (x2, y2), color=color, thickness=bbox_thickness)

            # Get class name
            class_name = self.class_map[int(class_label)] if self.class_map else str(class_label)

            # Draw class info
            text_size, _ = cv2.getTextSize(str(class_name), cv2.FONT_HERSHEY_SIMPLEX, text_scale, text_thickness)
            text_w, text_h = text_size
            visualize_image = cv2.rectangle(visualize_image, (x1, y1-5-text_h), (x1+text_w, y1), color=color, thickness=-1)
            visualize_image = cv2.putText(visualize_image, str(class_name), (x1, y1-5), cv2.FONT_HERSHEY_SIMPLEX, text_scale, (255, 255, 255), text_thickness)

        return visualize_image

    def visualize_by_plt(self, image):
        plt.imshow(image)
        plt.show()



import os
import time
from pathlib import Path
from typing import List

from loguru import logger

from netspresso.clients.tao import tao_client
from netspresso.enums.tao.action import ExperimentAction


class Experiment:
    def __init__(self, id, name, network_arch, token_handler) -> None:
        self.id = id
        self.name = name
        self.network_arch = network_arch
        self.token_handler = token_handler
        self.pretrained_model = None
        self.data = None
        self.train_specs = self.get_train_specs()
        self.export_specs = self.get_export_specs()
        self.evaluate_specs = self.get_evaluate_specs()
        self.job_map = {}
        self.train_job_cnt = 1
        self.export_job_cnt = 1

    def delete_experiment(self):
        try:
            logger.info("Deleting experiment...")
            response = tao_client.experiment.delete_experiment(
                self.token_handler.user_id, self.id, self.token_handler.headers
            )

            return response

        except Exception as e:
            logger.error(f"Delete experiment failed. Error: {e}")
            raise e

    def set_dataset(
        self, train_datasets: List[str], eval_dataset: str, inference_dataset: str, calibration_dataset: str
    ):
        self.data = {
            "train_datasets": train_datasets,
            "eval_dataset": eval_dataset,
            "inference_dataset": inference_dataset,
            "calibration_dataset": calibration_dataset,
        }

    def set_pretrained_model(self, pretrained_model_name: str):
        # Check if network architecture requires a pretrained model
        # Get experiments for the given network architecture
        experiments = tao_client.experiment.get_experiments(
            user_id=self.token_handler.user_id, headers=self.token_handler.headers, network_arch=self.network_arch
        )

        # Search for pretrained model with the given NGC path
        for experiment in experiments:
            ngc_path = experiment.get("ngc_path")
            if ngc_path and ngc_path.endswith(pretrained_model_name):
                print(ngc_path)
                logger.info(f"Pretrained model info: {experiment}")
                ptm_id = experiment.get("id")
                logger.info("Metadata for model with requested NGC Path")
                self.pretrained_model = {"base_experiment": [ptm_id]}
                break

        else:
            logger.warning("No pretrained model specified for the given PTM name.")

    def set_automl_with_bayesian(self, metric, automl_max_recommendations, override_automl_disabled_params, additional_automl_params, remove_default_automl_params):
        automl_information = {
            "automl_enabled": True,
            "automl_algorithm": "bayesian",
            "metric": metric,
            "automl_max_recommendations": automl_max_recommendations,
            "override_automl_disabled_params": override_automl_disabled_params,
            "automl_add_hyperparameters": str(additional_automl_params),
            "automl_remove_hyperparameters": str(remove_default_automl_params)
        }

        tao_client.experiment.partial_update_experiment(
            self.token_handler.user_id, self.id, automl_information, self.token_handler.headers
        )

    def set_automl_with_hyperband(self, metric, automl_R, automl_nu, epoch_multiplier, override_automl_disabled_params, additional_automl_params, remove_default_automl_params):
        automl_information = {
            "automl_enabled": True,
            "automl_algorithm": "hyperband",
            "metric": metric,
            "automl_R": automl_R,
            "automl_nu": automl_nu,
            "epoch_multiplier": epoch_multiplier,
            "override_automl_disabled_params": override_automl_disabled_params,
            "automl_add_hyperparameters": str(additional_automl_params),
            "automl_remove_hyperparameters": str(remove_default_automl_params)
        }

        tao_client.experiment.partial_update_experiment(
            self.token_handler.user_id, self.id, automl_information, self.token_handler.headers
        )

    def get_train_specs(self):
        try:
            logger.info("Getting train specs...")
            response = tao_client.experiment.get_specs_schema(
                self.token_handler.user_id, self.id, ExperimentAction.TRAIN, self.token_handler.headers
            )
            train_specs = response["default"]

            return train_specs

        except Exception as e:
            logger.error(f"Get train specs failed. Error: {e}")
            raise e

    def get_export_specs(self):
        try:
            logger.info("Getting export specs...")
            response = tao_client.experiment.get_specs_schema(
                self.token_handler.user_id, self.id, ExperimentAction.EXPORT, self.token_handler.headers
            )
            export_specs = response["default"]

            return export_specs

        except Exception as e:
            logger.error(f"Get export specs failed. Error: {e}")
            raise e

    def get_evaluate_specs(self):
        try:
            logger.info("Getting evaluate specs...")
            response = tao_client.experiment.get_specs_schema(
                self.token_handler.user_id, self.id, ExperimentAction.EVALUATE, self.token_handler.headers
            )
            evaluate_specs = response["default"]

            return evaluate_specs

        except Exception as e:
            logger.error(f"Get evaluate specs failed. Error: {e}")
            raise e

    def get_prune_specs(self):
        try:
            logger.info("Getting prune specs...")
            response = tao_client.experiment.get_specs_schema(
                self.token_handler.user_id, self.id, ExperimentAction.PRUNE, self.token_handler.headers
            )
            prune_specs = response["default"]

            return prune_specs

        except Exception as e:
            logger.error(f"Get prune specs failed. Error: {e}")
            raise e

    def get_retrain_specs(self):
        try:
            logger.info("Getting retrain specs...")
            response = tao_client.experiment.get_specs_schema(
                self.token_handler.user_id, self.id, ExperimentAction.RETRAIN, self.token_handler.headers
            )
            retrain_specs = response["default"]

            return retrain_specs

        except Exception as e:
            logger.error(f"Get retrain specs failed. Error: {e}")
            raise e

    def get_trt_engine_spces(self):
        try:
            logger.info("Getting trt engine specs...")
            response = tao_client.experiment.get_specs_schema(
                self.token_handler.user_id, self.id, ExperimentAction.GEN_TRT_ENGINE, self.token_handler.headers
            )
            trt_engine_specs = response["default"]

            return trt_engine_specs

        except Exception as e:
            logger.error(f"Get trt engine specs failed. Error: {e}")
            raise e

    def get_inference_spces(self):
        try:
            logger.info("Getting inference specs...")
            response = tao_client.experiment.get_specs_schema(
                self.token_handler.user_id, self.id, ExperimentAction.INFERENCE, self.token_handler.headers
            )
            inference_specs = response["default"]

            return inference_specs

        except Exception as e:
            logger.error(f"Get inference specs failed. Error: {e}")
            raise e

    def train(self, name: str, parent_job_id: str = None):
        try:
            logger.info("Running train...")
            update_data = {**self.data, **self.pretrained_model}
            tao_client.experiment.partial_update_experiment(
                self.token_handler.user_id, self.id, update_data, self.token_handler.headers
            )

            data = {
                "name": name,
                "action": ExperimentAction.TRAIN,
                "parent_job_id": parent_job_id,
                "specs": self.train_specs,
            }
            train_job_id = tao_client.experiment.run_experiment_jobs(
                self.token_handler.user_id, self.id, data, self.token_handler.headers
            )
            job_key = f"train_job_{self.train_job_cnt}"
            self.job_map[job_key] = train_job_id
            self.train_job_cnt += 1

            return train_job_id

        except Exception as e:
            logger.error(f"Train failed. Error: {e}")
            raise e

    def evaluate(self, parent_job_id):
        try:
            logger.info("Evaluating...")
            data = {
                "parent_job_id": parent_job_id,
                "action": ExperimentAction.EVALUATE,
                "specs": self.eval_specs,
            }
            evaluate_job_id = tao_client.experiment.run_experiment_jobs(
                self.token_handler.user_id, self.id, data, self.token_handler.headers
            )

            return evaluate_job_id

        except Exception as e:
            logger.error(f"Evaluate failed. Error: {e}")
            raise e

    def export(self, parent_job_id):
        try:
            logger.info("Exporting...")
            data = {
                "name": f"{self.name} export",
                "parent_job_id": parent_job_id,
                "action": ExperimentAction.EXPORT,
                "specs": self.export_specs,
            }
            export_job_id = tao_client.experiment.run_experiment_jobs(
                self.token_handler.user_id, self.id, data, self.token_handler.headers
            )
            job_key = f"export_job_{self.export_job_cnt}"
            self.job_map[job_key] = export_job_id
            self.export_job_cnt += 1

            return export_job_id

        except Exception as e:
            logger.error(f"Export failed. Error: {e}")
            raise e

    def prune(self, parent_job_id, prune_specs):
        try:
            logger.info("Pruning...")
            data = {
                "name": f"{self.name} prune",
                "parent_job_id": parent_job_id,
                "action": ExperimentAction.PRUNE,
                "specs": prune_specs,
            }
            prune_job_id = tao_client.experiment.run_experiment_jobs(
                self.token_handler.user_id, self.id, data, self.token_handler.headers
            )

            return prune_job_id

        except Exception as e:
            logger.error(f"Prune failed. Error: {e}")
            raise e

    def retrain(self, name: str, parent_job_id: str, retrain_specs):
        try:
            logger.info("Running train...")
            update_data = {**self.data, **self.pretrained_model}
            tao_client.experiment.partial_update_experiment(
                self.token_handler.user_id, self.id, update_data, self.token_handler.headers
            )

            data = {
                "name": name,
                "action": ExperimentAction.RETRAIN,
                "parent_job_id": parent_job_id,
                "specs": retrain_specs,
            }
            retrain_job_id = tao_client.experiment.run_experiment_jobs(
                self.token_handler.user_id, self.id, data, self.token_handler.headers
            )

            return retrain_job_id

        except Exception as e:
            logger.error(f"Retrain failed. Error: {e}")
            raise e

    def gen_trt_engine(self, parent_job_id, trt_engine_specs):
        try:
            logger.info("Generating TRT engine...")
            data = {
                "name": f"{self.name} trt engine",
                "parent_job_id": parent_job_id,
                "action": ExperimentAction.GEN_TRT_ENGINE,
                "specs": trt_engine_specs,
            }
            gen_trt_job_id = tao_client.experiment.run_experiment_jobs(
                self.token_handler.user_id, self.id, data, self.token_handler.headers
            )

            return gen_trt_job_id

        except Exception as e:
            logger.error(f"Generate TRT engine failed. Error: {e}")
            raise e

    def inference(self, parent_job_id, inference_specs):
        try:
            logger.info("Inferring...")
            data = {
                "name": f"{self.name} inference",
                "parent_job_id": parent_job_id,
                "action": ExperimentAction.INFERENCE,
                "specs": inference_specs,
            }
            inference_job_id = tao_client.experiment.run_experiment_jobs(
                self.token_handler.user_id, self.id, data, self.token_handler.headers
            )

            return inference_job_id

        except Exception as e:
            logger.error(f"Inference failed. Error: {e}")
            raise e

    def monitor_job_status(self, job_id, interval=15):
        try:
            logger.info("Monitoring experiment job stauts...")

            while True:
                response = tao_client.experiment.get_experiment_job(
                    self.token_handler.user_id, self.id, job_id, self.token_handler.headers
                )
                logger.info(response)
                if response.get("status") in ["Done", "Error", "Canceled"]:
                    break
                time.sleep(interval)

        except Exception as e:
            logger.error(f"Monitor experiment job staus failed. Error: {e}")
            raise e

        except KeyboardInterrupt:
            logger.info("End monitoring.")

    def get_experiment_jobs(self):
        try:
            logger.info("Getting experiment jobs...")
            response = tao_client.experiment.get_experiment_jobs(
                self.token_handler.user_id, self.id, self.token_handler.headers
            )

            return response

        except Exception as e:
            logger.error(f"Get experiment jobs failed. Error: {e}")
            raise e

    def get_experiment_job(self, job_id):
        try:
            logger.info("Getting experiment job...")
            response = tao_client.experiment.get_experiment_job(
                self.token_handler.user_id, self.id, job_id, self.token_handler.headers
            )

            return response

        except Exception as e:
            logger.error(f"Get experiment job failed. Error: {e}")
            raise e

    def delete_experiment_job(self, job_id):
        try:
            logger.info("Deleting experiment job...")
            response = tao_client.experiment.delete_experiment_job(
                self.token_handler.user_id, self.id, job_id, self.token_handler.headers
            )

            return response

        except Exception as e:
            logger.error(f"Delete experiment job failed. Error: {e}")
            raise e

    def resume_experiment_job(self, job_id):
        try:
            logger.info("Resuming experiment job...")
            response = tao_client.experiment.resume_experiment_job(
                self.token_handler.user_id, self.id, job_id, self.token_handler.headers
            )

            return response

        except Exception as e:
            logger.error(f"Resume experiment job failed. Error: {e}")
            raise e

    def cancel_experiment_job(self, job_id):
        try:
            logger.info("Canceling experiment job...")
            response = tao_client.experiment.cancel_experiment_job(
                self.token_handler.user_id, self.id, job_id, self.token_handler.headers
            )

            return response

        except Exception as e:
            logger.error(f"Cancel experiment job failed. Error: {e}")
            raise e

    def download_artifacts(self, job_id, output_dir, best_model=True, latest_model=False):
        try:
            logger.info("Downloading selective artifacts...")
            file_lists = tao_client.experiment.get_list_files(
                self.token_handler.user_id, self.id, job_id, self.token_handler.headers
            )
            logger.info(f"File lists: {file_lists}")

            # Save
            temptar = f"{job_id}.tar.gz"
            with tao_client.experiment.download_selective_files(
                self.token_handler.user_id,
                self.id,
                job_id,
                file_lists,
                best_model,
                latest_model,
                self.token_handler.headers,
            ) as r:
                r.raise_for_status()
                with open(temptar, "wb") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)

            # Untar to destination
            logger.info("Untarring...")
            Path(output_dir).mkdir(parents=True, exist_ok=True)
            tar_command = f"tar -xvf {temptar} -C {output_dir}/"
            os.system(tar_command)
            os.remove(temptar)

            return output_dir

        except Exception as e:
            logger.error(f"Download artifacts failed. Error: {e}")
            raise e



from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Union


@dataclass
class LoggingConfig:
    project_id: Optional[str] = None
    output_dir: Union[Path, str] = "./outputs"
    tensorboard: bool = True
    image: bool = True
    stdout: bool = True
    save_optimizer_state: bool = True
    onnx_input_size: List = field(default_factory=lambda: [512, 512])
    validation_epoch: int = 10
    save_checkpoint_epoch: Optional[int] = None

    def __post_init__(self):
        if self.save_checkpoint_epoch is None:
            self.save_checkpoint_epoch = self.validation_epoch



from .compression import (
    CompressionMethod,
    GroupPolicy,
    LayerNorm,
    Policy,
    RecommendationMethod,
    StepOp,
)
from .config import EndPointProperty, EnvironmentType, ServiceModule, ServiceName
from .credit import MembershipType, ServiceCredit, ServiceTask
from .device import (
    DeviceName,
    DisplaySoftwareVersion,
    HardwareType,
    SoftwareVersion,
    TaskStatus,
)
from .inference import Runtime
from .metadata import Status, TaskType
from .model import DataType, Extension, Framework, OriginFrom
from .module import Module
from .tao.action import ConvertAction, ExperimentAction
from .task import LauncherTask, Task, TaskStatusForDisplay
from .train import Optimizer, Scheduler

__all__ = [
    "ServiceCredit",
    "ServiceTask",
    "TaskType",
    "Status",
    "CompressionMethod",
    "RecommendationMethod",
    "Policy",
    "GroupPolicy",
    "LayerNorm",
    "Task",
    "Framework",
    "Framework",
    "Extension",
    "OriginFrom",
    "DataType",
    "DeviceName",
    "SoftwareVersion",
    "DisplaySoftwareVersion",
    "HardwareType",
    "TaskStatus",
    "Module",
    "ConvertAction",
    "ExperimentAction",
    "StepOp",
    "MembershipType",
    "DisplaySoftwareVersion",
    "LauncherTask",
    "TaskStatusForDisplay",
    "EnvironmentType",
    "ServiceModule",
    "EndPointProperty",
    "ServiceName",
    "Optimizer",
    "Scheduler",
    "Runtime",
]



from dataclasses import dataclass, field
from typing import List, Optional

from netspresso.metadata.common import InputShape, ModelInfo


@dataclass
class InputLayer:
    name: Optional[str] = None
    batch: Optional[int] = None
    channel: Optional[int] = None
    dimension: Optional[list] = None

    def __post_init__(self):
        if self.batch is None:
            self.batch = 1

    def to(self) -> InputShape:
        input_shape = InputShape()
        input_shape.batch = self.batch
        input_shape.channel = self.channel
        input_shape.dimension = self.dimension
        return input_shape


@dataclass
class ModelDetail:
    data_type: Optional[str] = None
    ai_model_format: Optional[str] = None
    framework: Optional[str] = None
    flops: Optional[float] = 0
    trainable_parameters: Optional[int] = 0
    non_trainable_parameters: Optional[int] = 0
    number_of_layers: Optional[int] = None
    graph_info: Optional[dict] = None
    input_layers: Optional[List[InputLayer]] = field(default_factory=list)

    def __post_init__(self):
        if self.input_layers:
            self.input_layers = [InputLayer(**input_layer) for input_layer in self.input_layers]

        if self.trainable_parameters is None:
            self.trainable_parameters = 0
        if self.non_trainable_parameters is None:
            self.non_trainable_parameters = 0


@dataclass
class ModelStatus:
    is_deleted: Optional[bool] = False
    is_convertible: Optional[bool] = False
    is_compressible: Optional[bool] = False
    is_benchmarkable: Optional[bool] = False
    is_uploaded: Optional[bool] = False


@dataclass
class ModelBase:
    user_id: str
    ai_model_id: str
    uploaded_file_name: Optional[str]
    file_size_in_mb: Optional[float]
    md5_checksum: Optional[str]
    bucket_name: Optional[str]
    object_path: Optional[str]
    task: str
    display_name: Optional[str]
    error_log: Optional[str]
    status: Optional[ModelStatus]
    detail: Optional[ModelDetail]

    def __post_init__(self):
        self.status = ModelStatus(**self.status)
        self.detail = ModelDetail(**self.detail)

    def to(self) -> ModelInfo:
        model_info = ModelInfo()
        model_info.data_type = self.detail.data_type
        model_info.framework = self.detail.framework
        model_info.input_shapes = [
            InputShape(
                batch=input_layer.batch,
                channel=input_layer.channel,
                dimension=input_layer.dimension
            ) for input_layer in self.detail.input_layers
        ]
        return model_info



from functools import partial

import numpy as np


def anchor_free_decoupled_head_decode(pred, original_shape, score_thresh=0.7):
    pred = pred['pred']
    stage_strides= [original_shape[-1] // o.shape[-1] for o in pred]

    hw = [x.shape[-2:] for x in pred]
    dim_len = pred[0].shape[1]

    pred = np.concatenate([x.reshape(1, dim_len, -1) for x in pred], axis=2).transpose(0, 2, 1)
    pred[..., 4:] = 1 / (1 + (np.exp(-pred[..., 4:])))

    grids = []
    strides = []
    for (hsize, wsize), stride in zip(hw, stage_strides):
        yv, xv = np.meshgrid(np.arange(hsize, dtype='float32'), np.arange(wsize, dtype='float32'), indexing='ij')
        grid = np.stack((xv, yv), 2).reshape(1, -1, 2)
        grids.append(grid)
        shape = grid.shape[:2]
        strides.append(np.full((*shape, 1), stride, dtype='float32'))

    grids = np.concatenate(grids, axis=1)
    strides = np.concatenate(strides, axis=1)

    pred = np.concatenate([
        (pred[..., 0:2] + grids) * strides,
        np.exp(pred[..., 2:4]) * strides,
        pred[..., 4:]
    ], axis=-1)

    box_corner = np.empty(pred.shape)
    box_corner[:, :, 0] = pred[:, :, 0] - pred[:, :, 2] / 2
    box_corner[:, :, 1] = pred[:, :, 1] - pred[:, :, 3] / 2
    box_corner[:, :, 2] = pred[:, :, 0] + pred[:, :, 2] / 2
    box_corner[:, :, 3] = pred[:, :, 1] + pred[:, :, 3] / 2
    pred[:, :, :4] = box_corner[:, :, :4]

    # Discard boxes with low score
    detections = []
    for p in pred:
        class_pred = np.argmax(p[:, 5:], 1, keepdims=True)
        class_conf = p[np.arange(p.shape[0]), 5+class_pred.squeeze()][:, np.newaxis]

        conf_mask = (p[:, 4] * class_conf.squeeze() >= score_thresh).squeeze()

        # x1, y1, x2, y2, obj_conf, pred_score, pred_label
        detections.append(
            np.concatenate((p[:, :5], class_conf, class_pred), axis=1)[conf_mask]
        )

    return detections

def nms_fast_rcnn(dets, scores, thresh):
    '''
    dets is a numpy array : num_dets, 4
    scores ia  nump array : num_dets,
    '''
    x1 = dets[:, 0]
    y1 = dets[:, 1]
    x2 = dets[:, 2]
    y2 = dets[:, 3]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1] # get boxes with more ious first

    keep = []
    while order.size > 0:
        i = order[0] # pick maxmum iou box
        keep.append(i)
        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        w = np.maximum(0.0, xx2 - xx1 + 1) # maximum width
        h = np.maximum(0.0, yy2 - yy1 + 1) # maxiumum height
        inter = w * h
        ovr = inter / (areas[i] + areas[order[1:]] - inter)

        inds = np.where(ovr <= thresh)[0]
        order = order[inds + 1]

    return keep

def nms(prediction, nms_thresh=0.45):
    output = [np.zeros((0, 7)) for _ in range(len(prediction))]
    for i, image_pred in enumerate(prediction):

        # If none are remaining => process next image
        if not image_pred.shape[0]:
            continue

        nms_out_index = nms_fast_rcnn(
            image_pred[:, :4],
            image_pred[:, 4] * image_pred[:, 5],
            nms_thresh,
        )

        image_pred = image_pred[nms_out_index]
        output[i] = np.concatenate((output[i], image_pred))

    return output


class DetectionPostprocessor:
    def __init__(self, score_thresh, nms_thresh):
        self.decode_outputs = partial(anchor_free_decoupled_head_decode, score_thresh=score_thresh)
        self.postprocess = partial(nms, nms_thresh=nms_thresh)

    def __call__(self, outputs, original_shape):
        pred = outputs

        if self.decode_outputs:
            pred = self.decode_outputs(pred, original_shape)
        if self.postprocess:
            pred = self.postprocess(pred)

        pred = [(np.concatenate([p[:, :4], p[:, 4:5] * p[:, 5:6]], axis=-1),
                      p[:, 6].astype('int'))
                      for p in pred]
        return pred



import dataclasses
from abc import ABC, abstractmethod

from netspresso.clients.launcher.v2.schemas import ResponseItem, ResponseItems


class ModelInterface(ABC):
    @abstractmethod
    def get_upload_url(
        self, request_params: dataclasses, headers: dataclasses
    ) -> ResponseItem:
        """
        :param request_params:
        :param headers:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def get_download_url(self, headers: dataclasses, ai_model_id: str) -> str:
        """
        :param headers:
        :param ai_model_id:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def upload(
        self, request_body: dataclasses, file: dataclasses, headers: dataclasses
    ) -> str:
        """
        :param request_body:
        :param file:
        :param headers:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def read(
        self, request_params: dataclasses, headers: dataclasses, ai_model_id: str
    ) -> ResponseItem:
        """
        :param request_params:
        :param headers:
        :param ai_model_id:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def read_all(
        self, request_params: dataclasses, headers: dataclasses
    ) -> ResponseItems:
        """
        :param request_params:
        :param headers:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def delete(self, headers: dataclasses, ai_model_id: str) -> ResponseItem:
        """
        :param headers:
        :param ai_model_id:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def validate(self, request_body: dataclasses, headers: dataclasses) -> ResponseItem:
        """
        :param request_body:
        :param headers:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def status(self, headers: dataclasses, ai_model_id: str) -> ResponseItem:
        """
        :param headers:
        :param ai_model_id:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def options(self, headers: dataclasses, ai_model_id: str) -> ResponseItems:
        """
        :param headers:
        :param ai_model_id:
        :return:
        """
        raise NotImplementedError



from dataclasses import dataclass, field

from netspresso.clients.auth.response_body import CreditResponse
from netspresso.clients.auth.v2.schemas.common import AbstractResponse


@dataclass
class CreditWithType:
    user_id: str
    free_credit: int
    paid_credit: int
    total_credit: int


@dataclass
class SummarizedCreditResponse(AbstractResponse):
    data: CreditWithType = field(default_factory=CreditWithType)

    def __post_init__(self):
        self.data = CreditWithType(**self.data)

    def to(self) -> CreditResponse:
        return CreditResponse(
            free=self.data.free_credit,
            reward=0,
            contract=0,
            paid=self.data.paid_credit,
            total=self.data.total_credit,
        )
